/*
 * chk4mail - mail folder check
 * $Id: chk4mail.c,v 1.11 2004/11/14 16:55:19 johans Exp $
 */

/*
 * Copyright (C) 1998 - 2004, Johan van Selst <johans@stack.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <string.h>
#include <dirent.h>
#include <utime.h>
#include <err.h>
#include <sys/utsname.h>
#include <paths.h>

#include "config.h"
#ifdef		HAVE_GETOPT_H
#include <getopt.h>
#endif		/* HAVE_GETOPT_H */
#ifndef		HAVE_STRLCAT
#include "strlcat.h"
#endif		/* HAVE_STRLCAT */
#ifndef		HAVE_STRLCPY
#include "strlcpy.h"
#endif		/* HAVE_STRLCPY */

#define MAXHEAD		256
#define	MAXFRM		10240
#define MAXDIR		1024
#define MAXFILE		80
#define SUBMAILDIR	1

int			spc = 5, hideold = 0, quiet = 0, brief = 0, oflag = 0,
			oanswer = 0, unreadasold = 0, printfrm = 0, touch = 0;
u_int		mlen = 4;
static char	rcsid[] = "$Id: chk4mail.c,v 1.11 2004/11/14 16:55:19 johans Exp $";

typedef struct		summary
{
	unsigned int		new;
	unsigned int		flag;
	unsigned int		answer;
	unsigned int		total;
	char				*frm;
	unsigned int		frmsz;
} summary_t;

void
usage(void)
{
	/* give clue */
	fputs("chk4mail " PACKAGE_VERSION " - mail folder check\n"
		"Mbox, MMDF, MH, Maildir and Extended Maildir folders are recognized.\n"
		"\n"
		"-# <n>     Pad the number of messages with n spaces\n"
		"-a         Show how many messages have been answered\n"
		"-c         Include folder names with capitals\n"
		"-d <dir>   Look for mailfolders in this directory instead\n"
		"-f         Show how many messages have been flagged as important\n"
		"-h         Hide folders containing only old mail\n"
		"-i         Only parse standard incoming mailfolder\n"
		"-l         Always follow symbolic links\n"
		"-m <n>     Set max folder name lenght (for layout padding)\n"
		"-n         Only parse folders that received new mail\n"
		"-q         Suppress errors about missing or unreadable files\n"
		"-r <host>  Execute the command on an other host (with rsh)\n"
		"-s         Short output shows only files with new mail, implies -n\n"
		"-t         Mark old files with new messages as unread ('touch')\n"
		"-u         Don't treat unread messages as new mail\n"
		"-x         Include folder names with leading dots\n"
		"Folders may be specified as arguments as well (filename or =folder)\n"
		"Read the manual for more details.\n",
		stderr);
	exit(1);
}

void
fail(const char *msg, const char *file)
{
	/* write friendly message */
	errx(1, "%s: %s\n", msg, file);
}

int
parse_mbox_mmdf(FILE *file, summary_t *sum, const int mmdf)
{
	int			nrs, isread, header, isnew;
	long		skip;
	static char	line[MAXHEAD],
				myfrom[MAXHEAD],
				mysubj[MAXHEAD],
				dummy[MAXHEAD];
	char		*temp;

	nrs = isread = 0;
	while (!feof(file))
	{
		skip = header = 0;
		isnew = 1;

		/* parse headers */
		while (fgets(line, MAXHEAD, file) &&
				*line != '\r' && *line != '\n')
		{
			/* look for begin of a new message */
			if ((!mmdf && !strncmp(line, "From ", 5)) ||
					(mmdf && !strncmp(line, "\1\1\1\1", 4)))
				header = 1;
			else if (header)
			{
				if (!strncmp(line, "Content-Length: ", 16))
					skip = atol(line + 16);
				else if (!strncmp(line, "Status: RO", 10) ||
						!strncmp(line, "Status: OR", 10))
				{
					isread++;
					isnew = 0;
				}
				else if (unreadasold && !strncmp(line, "Status: O", 9))
				{
					isread++;
					isnew = 0;
				}
				else if (!strncmp(line, "X-Status: ", 10))
				{
					if (strchr(line + 10, 'F'))
						sum->flag++;
					if (strchr(line + 10, 'A'))
						sum->answer++;
				}
				/* nfrm compatibility: guess sender name */
				else if (printfrm && !strncmp(line, "From: ", 6))
				{
					if (sscanf(line + 6, "\"%[^\"]\" <%s>\n", myfrom,dummy)<2 &&
						sscanf(line + 6, "<%s> \"%[^\"]\"\n", dummy,myfrom)<2 &&
						sscanf(line + 6, "%[^<] <%s>\n", myfrom, dummy)<2 &&
						sscanf(line + 6, "<%s> %[^<]\n", dummy, myfrom)<2 &&
						sscanf(line + 6, "(%[^)]) %s\n", myfrom, dummy)<2 &&
						sscanf(line + 6, "%s (%[^)])\n", dummy, myfrom)<2)
					{
						strlcpy(myfrom, line + 6, MAXHEAD);
						if ((temp = strchr(myfrom, '\n')))
							*temp = '\0';
					}
				}
				/* nfrm compatibility: copy subject */
				else if (printfrm && !strncmp(line, "Subject: ", 9))
				{
					strlcpy(mysubj, line + 9, MAXHEAD);
					if ((temp = strchr(mysubj, '\n')))
						*temp = '\0';
				}
			}
		}

		/* skip body (hopefully) */
		if (header)
		{
			nrs++;
			fseek(file, skip, SEEK_CUR);
			/* nfrm compatibility output */
			if ((2 == printfrm) || (printfrm && isnew))
			{
				if (strlen(sum->frm) + 80 > sum->frmsz)
				{
						sum->frmsz *= 2;
						if (!(temp = realloc(sum->frm, sum->frmsz)))
								fail("Out of memory", "realloc()");
						sum->frm = temp;
				}
				temp = sum->frm + strlen(sum->frm);
				sprintf(temp, "%-24.24s %-53.53s\n", myfrom, mysubj);
			}
		}
	}

	if (mmdf)
		nrs--;

	sum->total = nrs;
	sum->new = nrs >= isread ? nrs - isread : 0;

	return 0;
}

int
parse_maildir(const char *subdir, summary_t *sum)
{
	char			*filename, *p;
	struct dirent	*file;
	DIR				*dir;
	int				isread;

	sum->new = 0;
	sum->total = 0;

	asprintf(&filename, "%s/cur", subdir);
	if (!(dir = opendir(filename)))
		fail("Unable to open subdirectory", filename);

	while ((file = readdir(dir)))
	{
		if (!strcmp(file->d_name, ".") || !strcmp(file->d_name, ".."))
			continue;
		isread = 0;
		if ((p = strrchr(file->d_name, ':')) && !strncmp(p, ":2,", 3))
		{
			p += 2;
			while (*(++p))
				if ('S' == *p)
					isread = 1;
				else if (oanswer && 'R' == *p)
					(sum->answer)++;
				else if (oflag && 'F' == *p)
					(sum->flag)++;
		}
		if (isread || unreadasold)
			(sum->total)++;
		else
			(sum->new)++;
	}
	closedir(dir);

	sprintf(filename, "%s/new", subdir);
	if (!(dir = opendir(filename)))
		fail("Unable to open subdirectory", filename);

	while ((file = readdir(dir)))
		if (strcmp(file->d_name, ".") && strcmp(file->d_name, ".."))
			(sum->new)++;
	closedir(dir);

	sum->total += sum->new;
	free(filename);
	return 0;
}

int
parse_mh(const char *subdir, summary_t *sum)
{
	DIR				*dir;
	struct dirent	*file;
	struct stat		statbuf;
	const u_int		len = strlen(subdir) + 7;
	char			*x, *filename;
	char			real;

	if (!(dir = opendir(subdir)))
		fail("Unable to open subdirectory", subdir);

	filename = malloc(len);
	memset(filename, 0, len);
	/* total = #filenames consisting of numbers */
	for (sum->total = 0; (file = readdir(dir)); )
	{
		for (real = 1, x = file->d_name; *x; x++)
			real &= *x >= '0' && *x <= '9';
		if (real)
		{
			(sum->total)++;
			snprintf(filename, len, "%s/%s", subdir, file->d_name);
			if (!stat(filename, &statbuf))
				if (statbuf.st_mtime >= statbuf.st_atime)
					(sum->new)++;
		}
	}

	closedir(dir);
	free(filename);
	return 0;
}

int
parse_dir(const char *dir, summary_t *sum)
{
	struct stat		filestat;
	char			*filename;

	asprintf(&filename, "%s/cur", dir);
	if (lstat(filename, &filestat) < 0 || !S_ISDIR(filestat.st_mode))
	{
		/* parse MH-style mailfolder */
		free(filename);
		return parse_mh(dir, sum);
	}
	else
	{
		/* parse cur/new/tmp-style mailfolder */
		free(filename);
		return parse_maildir(dir, sum);
	}
}

int
desc(char *buf, const unsigned int i, const char *msg)
{
	if (i <= 0)
		return sprintf(buf, "%*c %s", spc, i ? '?' : '-', msg);
	else
		return sprintf(buf, "%*u %s", spc, i, msg);
}

int
display(const char *folder, const summary_t *sum)
{
	static char	line[256];
	char		*p = line;

	if (printfrm)
	{
		if (sum->new || (sum->total && (2 == printfrm)))
			puts(sum->frm);
		else
			fprintf(stdout, "%s has no new mail\n", folder);
		return 0;
	}
	else if ((sum->total > 0) && (sum->new || !hideold))
	{
		p += sprintf(p, "%-*s", mlen, folder);
		p += desc(p, sum->new, "new");
		if (oflag)
			p += desc(p, sum->flag, "flag");
		if (oanswer)
			p += desc(p, sum->answer, "answ");
		fprintf(stdout, "%s of total%*d\n", line, spc, sum->total);
	}
	else if (brief && sum->new > 0)
		fprintf(stdout, "%s\n", folder);

	return 0;
}

int
parse_folder(const char *dir, const char *filenm, const int flink)
{
	struct stat		filestat;
	struct utimbuf	filetime;
	summary_t		sum = { 0, 0, 0, 0, NULL, 0 };
	char			*filename;
	FILE			*file;
	char			c;

	if (printfrm)
	{
		sum.frmsz = MAXFRM;
		sum.frm = malloc(MAXFRM);
		sum.frm[0] = '\0';
	}

	asprintf(&filename, "%s/%s", dir, filenm);

	if (lstat(filename, &filestat) < 0)
	{
		if (quiet)
			return 2;
		else
			fail("Unable to stat file", filename);
	}

	/* Follow symlinks iff 'flink' */
	if (S_ISLNK(filestat.st_mode) && flink &&
			(stat(filename, &filestat) < 0))
	{
		if (quiet)
			return 2;
		else
			fail("Unable to stat file", filename);
	}
	if (S_ISDIR(filestat.st_mode))
	{
#ifdef		SUBMAILDIR
		char			*subsub = malloc(MAXFILE), *p;
		char			dirlist[MAXDIR][MAXFILE];
		DIR				*subdir;
		struct dirent	*subfile;
		int				max, oldmlen;
#endif		/* SUBMAILDIR */

		/* parse directory mailfolder */
		parse_dir(filename, &sum);
		display(filenm, &sum);

#ifdef		SUBMAILDIR
		if (!(subdir = opendir(filename)))
			fail("Unable to open subdirectory", filename);

		/* sort maildir */
		max = 0;
		oldmlen = mlen;
		memset(dirlist, 0, sizeof(dirlist));
		while ((subfile = readdir(subdir)))
		{
			size_t	len;

			/* check if this mailfolder is worth reading */
			if (subfile->d_type != DT_DIR ||
					!strcmp(subfile->d_name, ".") ||
					!strcmp(subfile->d_name, ".."))
				continue;

			len = strlen(subfile->d_name) + strlen(filenm) + 1;
			if (len > mlen)
				mlen = len;
			strncpy(dirlist[max++], subfile->d_name, MAXFILE);
		}
		closedir(subdir);

		qsort(dirlist, MAXDIR, MAXFILE,
			(int (*)(const void *, const void *))strcmp);

		/* parse maildir */
		for (max = MAXDIR - max; max < MAXDIR; max++)
		{
			snprintf(subsub, MAXFILE, "%s/%s/new", filename, dirlist[max]);
			if ((stat(subsub, &filestat) < 0) || !S_ISDIR(filestat.st_mode))
				continue;

			snprintf(subsub, MAXFILE, "%s/%s", filename, dirlist[max]);
			parse_maildir(subsub, &sum);
			snprintf(subsub, MAXFILE, "%s/%s",
				filenm,
				'.' == dirlist[max][0] ? &dirlist[max][1] : dirlist[max]);

			for (p = subsub + strlen(filenm); *p; p++)
				if ('.' == *p)
					*p = '/';
			display(subsub, &sum);
		}
		mlen = oldmlen;
#endif		/* SUBMAILDIR */

		free(filename);
		return 1;
	}
	if (!S_ISREG(filestat.st_mode))
		return 2;
	if (!filestat.st_size)
		return 3;

	/* open and parse mailfolder */
	filetime.actime = filestat.st_atime;
	filetime.modtime = filestat.st_mtime;

	if (hideold > 1)
	{
		if (brief || filetime.modtime < filetime.actime)
		{
			free(filename);
			return 0;
		}
	}

	if (!(file = fopen(filename, "r")))
	{
		if (quiet)
			return 4;
		else
			fail("Unable to open file", filename);
		/* NOTREACHED */
	}

	c = fgetc(file);
	ungetc(c, file);
	parse_mbox_mmdf(file, &sum, '\1' == c);
	display(filenm, &sum);

	/* close and preserve timestamp */
	fclose(file);
	if (touch && filetime.modtime < filetime.actime && sum.new)
		filetime.modtime = time(NULL);
	utime(filename, &filetime);
	free(filename);

	return 0;
}

int
main(int argc, char **argv)
{
	char			*maildir = NULL, *inbox = NULL, *remote = NULL;
	DIR				*dir;
	struct dirent	*file;
	char			*slash;
	char			dirlist[MAXDIR][MAXFILE];
	struct utsname	uts;
	int				opt, max,
					inonly = 0, flink = 0, caps = 0, optc = 0, showdot = 0;
	u_int			len;
	char			**ap, **optv, *options;

	len = strlen(argv[0]);
	/* nfrm copatibility */
	if ((len > 5 && !strcmp(argv[0] + len - 5, "/nfrm")) ||
			(4 == len && !strcmp(argv[0], "nfrm")))
		inonly = printfrm = 1;
	else if ((len > 4 && !strcmp(argv[0] + len - 4, "/frm")) ||
			(3 == len && !strcmp(argv[0], "frm")))
		inonly = printfrm = 2;

	/* this is a dirty hack */
	if (getenv("MAIL_UNREAD_AS_OLD"))
			unreadasold = 1;

	/* be nice to lazy people - insert defaults before cmdline */
	/* TODO: maybe put this in a configuration file some day? */
	if ((options = getenv("CHK4MAIL_OPTIONS")))
	{
		int i;

		/* optv = { argv[0], options[0..], argv[1..] } */
		optv = (char **)malloc(argc + strlen(options));
		optv[0] = argv[0];
		ap = &optv[1];
		for (optc = 1; (*ap++ = strsep(&options, " \t")); optc++)
			/* DO NOTHING */;

		for (i = 1; i < argc; i++)
			optv[optc++] = argv[i];
	}
	else
	{
		optc = argc;
		optv = argv;
	}

	while ((opt = getopt(optc, optv, "#:acd:fhilm:nqr:stux")) != EOF)
		switch (opt)
		{
		case '#':
			if ((spc = atoi(optarg)) <= 0)
				usage();
			break;
		case 'a':
			oanswer = 1;
			break;
		case 'c':
			caps = 1;
			break;
		case 'd':
			maildir = strdup(optarg);
			break;
		case 'f':
			oflag = 1;
			break;
		case 'h':
			hideold = 1;
			break;
		case 'i':
			inonly = 1;
			break;
		case 'l':
			flink = 1;
			break;
		case 'm':
			if ((mlen = atoi(optarg)) <= 0)
				usage();
			break;
		case 'n':
			hideold = 2;
			break;
		case 'q':
			quiet = 1;
			break;
		case 'r':
			remote = strdup(optarg);
			break;
		case 's':
			brief = 1;
			break;
		case 't':
			touch = 1;
			break;
		case 'u':
			unreadasold = 1;
			break;
		case 'x':
			showdot = 1;
			break;
		default:
			usage();
			/* NOTREACHED */
		}

	if (brief)
		hideold = 2;
	/* run cmd on remote server */
	if (remote)
	{
		if (uname(&uts) < 0)
			fail("uname() failed", "???");
		if ((slash = strchr(uts.nodename, '.')))
			*slash = '\0';
		if (strcmp(remote, uts.nodename))
		{
			while (--argc)
				*(argv[argc]-1) = ' ';
			/* execl() does not return on success */
			if (execl("/usr/bin/rsh", remote, *argv, NULL) < 0)
				fail("execl() failed", remote);
		}
	}
	/* find inbox */
	if ((inbox = getenv("MAIL")))
		inbox = strdup(inbox);
	else
	{
		char	*user;

		if ((user = getenv("USER")))
			asprintf(&inbox, "%s/%s", _PATH_MAILDIR, user);
	}

	/* Shortcut to the end */
	if (inonly)
		goto INBOX;

	/* find maildir */
	if (!maildir)
	{
		asprintf(&maildir, "%s/Mail", getenv("HOME"));
		if (!(dir = opendir(maildir)))
			sprintf(maildir, "%s/mail", getenv("HOME"));
	}
	if (!(dir = opendir(maildir)))
		fail("Unable to open maildirectory", maildir);

	/* only parse specified mailboxes and exit */
	if (optind < argc)
	{
		/* for a nice layout: find max filename length first */
		opt = optind;
		while (opt < argc)
		{
			if ((slash = strrchr(argv[opt], '/')))
				len = strlen(slash + 1);
			else
				len = strlen(argv[opt]);
			opt++;
			if (len > mlen)
				mlen = len;
		}
		opt = optind;
		while (opt < argc)
		{
			if ('=' == *argv[opt] || '+' == *argv[opt])
				parse_folder(maildir, argv[opt] + 1, 1);
			else if ((slash = strrchr(argv[opt], '/')))
			{
				*slash = '\0';
				parse_folder(argv[opt], slash + 1, 1);
			}
			else
			{
				getcwd(maildir, MAXFILE);
				parse_folder(maildir, argv[opt], 1);
			}
			opt++;
		}
		closedir(dir);
		return 0;
	}

	/* sort maildir */
	max = 0;
	memset(dirlist, 0, sizeof(dirlist));
	while ((file = readdir(dir)))
	{
		/* check if this mailfolder is worth reading */
		len = strlen(file->d_name);
		if (showdot &&
				len &&
				'.' == *file->d_name &&
				file->d_name[1] != '\0' &&
				(file->d_name[2] != '\0' || file->d_name[1] != '.'))
			/* .text is fine */;
		else if (len &&
				(*file->d_name < 'a' || *file->d_name > 'z') &&
				(*file->d_name < 'A' || *file->d_name > 'Z' || !caps) &&
				(*file->d_name < '0' || *file->d_name > '9'))
			continue;

		if ((len > 4) && !strcmp((char *)(file->d_name + len-5), ".lock"))
			continue;

		if (len > mlen)
			mlen = len;
		strncpy(dirlist[max++], file->d_name, MAXFILE);
	}

	qsort(dirlist, MAXDIR, MAXFILE,
		(int (*)(const void *, const void *))strcmp);
	max = MAXDIR-max;

	/* parse maildir */
	while (max < MAXDIR)
		parse_folder(maildir, dirlist[max++], flink);
	closedir(dir);

INBOX:
	/* parse inbox */
	if (inbox && strlen(inbox))
	{
		if ((slash = strrchr(inbox, '/')))
			*slash = '\0';
		parse_folder(inbox, slash ? slash+1 : inbox, 1);
	}

	/* It's no unused variable... See? Here it is... */
	(void)rcsid;
	return 0;
}

/* vi:ts=4 sw=4
 */
